import os
import flask
import flask_cors
import threading
import webbrowser
from pathlib import Path
from ..api import API
from .editor import get_editor_bp
from .player import get_player_bp
from ..overloads import overloads

HOST = os.getenv("ABSTRA_HOST", "localhost")


def create_app(api: API):
    app = flask.Flask(__name__)
    flask_cors.CORS(app)

    editor = get_editor_bp(api)
    app.register_blueprint(editor, url_prefix="/_editor")

    player = get_player_bp(api)
    app.register_blueprint(player)

    return app


def serve(workspace_root: Path, port: int, debug, use_reloader, load_dotenv):
    os.environ["ABSTRA_SERVER"] = "true"
    overloads()

    api = API(workspace_root)

    credential = api.get_credentials()
    if credential:
        os.environ["ABSTRA_API_TOKEN"] = credential

    app = create_app(api)
    if debug:
        threading.Timer(
            1, lambda: webbrowser.open(f"http://{HOST}:{port}/_editor")
        ).start()
    app.run(
        host=HOST,
        port=port,
        debug=debug,
        use_reloader=use_reloader,
        load_dotenv=load_dotenv,
    )



import os, typing
from .sqlite import SqliteDB, TableNotFound  # expoted
from .legacy import *  # deprecated


def get_db(path: typing.Optional[str] = None):
    env_db_url = os.getenv("ABSTRA_DATABASE_URL")
    if path:
        db_url = path
    elif env_db_url:
        db_url = env_db_url
    else:
        db_url = "sqlite:///:memory:"

    if db_url.startswith("postgres://"):
        raise NotImplementedError("Postgres is not supported yet")
    elif db_url.startswith("mysql://"):
        raise NotImplementedError("MySQL is not supported yet")
    else:
        return SqliteDB(db_url)


def get_table(table_name: str):
    return get_db().get_table(table_name)



from ..widget_base import Output
from typing import Any
import json


class PandasOutput(Output):
    type = "pandas-output"

    def __init__(self, df: Any, **kwargs):
        self.df = df
        self.columns = kwargs.get("columns", 1)
        self.full_width = kwargs.get("full_width", False)
        self.display_index = kwargs.get("display_index", False)
        self.label = kwargs.get("label", None)
        self.actions = kwargs.get("actions", [])
        self.filterable = kwargs.get("filterable", False)
        self.editable = kwargs.get("editable", False)

    def json(self, **kwargs):
        return {
            "type": self.type,
            "table": json.loads(self.df.to_json(orient="table")),
            "columns": self.columns,
            "fullWidth": self.full_width,
            "displayIndex": self.display_index,
            "label": self.label,
            "actions": self.actions,
            "filterable": self.filterable,
            "editable": self.editable,
        }



import tempfile
import unittest
from uuid import uuid4 as uuid
from pathlib import Path

from abstra_server.api import API

from .fixtures import init_dir, clear_dir


class TestDashes(unittest.TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_api_create(self):
        api = API(root=self.root)

        dashes_len = len(api.get_dashes())

        dash = api.create_dash()

        self.assertEqual(len(api.get_dashes()), dashes_len + 1)

        self.assertEqual(dash.layout.slot.__dict__, {})

        file_path = Path(self.root, dash.file)

        self.assertTrue(file_path.exists())

    def test_api_update(self):
        api = API(root=self.root)

        dash = api.create_dash()

        new_layout = {
            "version": "0.2",
            "props": {},
            "slot": {
                "75838e71-92cd-4d8c-bfc0-f9ae23c32559": {
                    "events": {},
                    "id": "75838e71-92cd-4d8c-bfc0-f9ae23c32559",
                    "props": {"label": '"Click here!"'},
                    "position": {
                        "rowStart": 8,
                        "rowEnd": 9,
                        "colStart": 8,
                        "colEnd": 9,
                    },
                    "type": "click-input",
                    "variable": None,
                }
            },
        }

        updated_dash = api.update_dash(
            dash.path, dict(title="new title", layout=new_layout)
        )
        self.maxDiff = None
        self.assertEqual(updated_dash.title, "new title")
        self.assertEqual(updated_dash.layout.__dict__, new_layout)

    def test_updating_file_from_existing_to_non_existing(self):
        # given a dash
        api = API(root=self.root)
        dash = api.create_dash()
        old_file = Path(self.root, dash.file)
        new_file = Path(self.root, "non-existing-file.py")

        # when updating the file to a non existing file
        api.update_dash(dash.path, dict(file="non-existing-file.py"))

        # then the existing file is renamed
        self.assertFalse(old_file.exists())
        self.assertTrue(new_file.exists())

    def test_updating_file_from_non_existing_to_existing(self):
        # given a dash
        api = API(root=self.root)
        dash = api.create_dash()
        old_file = Path(self.root, dash.file)
        new_file = Path(self.root, "existing-file.py")
        old_file.unlink()
        new_file.write_text("print('hello')")

        # when updating the file to a non existing file
        api.update_dash(dash.path, dict(file="existing-file.py"))

        # then just the file property is updated
        self.assertFalse(old_file.exists())
        self.assertTrue(new_file.exists())


if __name__ == "__main__":
    unittest.main()



import builtins

from .generated.inputs import read
from .generated.outputs import display_plotly


def _overload_input():
    read_overload = lambda msg="": read(msg)
    builtins.input = read_overload
    builtins.raw_input = read_overload


# def _overload_print():
#     def print_overload(*values):
#         page = Page()
#         for value in values:
#             page.display(value)
#         return page.run()

#     builtins.print = print_overload


def _overload_plotly_show():
    try:
        from plotly.io import renderers
        from plotly.graph_objects import Figure
        from plotly.io._base_renderers import ExternalRenderer

        class FormsRenderer(ExternalRenderer):
            def render(self, fig_dict):
                fig = Figure(data=fig_dict)
                display_plotly(fig)
                return {}

        renderers["abstra-forms"] = FormsRenderer()
        renderers.default = "abstra-forms"
    except Exception as e:
        pass


def overload_to_widgets():
    _overload_input()
    _overload_plotly_show()



import os
from pathlib import Path
import unittest
from abstra_cli.utils import files_from_directory, get_ignore_files
from .fixtures import init_dir, clear_dir


class TestListing(unittest.TestCase):
    def add_file(self, name, content, path=None):
        filepath = (path or self.path).joinpath(name)
        filepath.write_text(content, encoding="utf-8")
        return filepath

    def add_folder(self, name, path=None):
        folderpath = (path or self.path).joinpath(name)
        folderpath.mkdir()
        return folderpath

    def setUp(self) -> None:
        self.path = init_dir()
        self.abstra_json_path = self.path.joinpath("abstra.json")

    def tearDown(self) -> None:
        clear_dir(self.path)

    def assertSamePaths(self, paths, expected):
        self.assertEqual(set([str(p) for p in paths]), set([str(p) for p in expected]))

    def test_empty_directory(self):
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [self.abstra_json_path])

    def test_no_ignore(self):
        filepath = self.add_file("foo", "bar")
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [filepath, self.abstra_json_path])

    def test_ignore_file(self):
        self.add_file(".abstraignore", "ignored")
        self.add_file("ignored", "foo")
        tracked = self.add_file("tracked", "bar")
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [tracked, self.abstra_json_path])

    def test_ignore_empty(self):
        self.add_folder("ignored")
        self.add_folder("empty")
        self.add_file(".abstraignore", "ignored")
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [self.abstra_json_path])

    def test_ignore_folder(self):
        ignored = self.add_folder("ignored")
        self.add_file("abc", "foo", path=ignored)
        ignored2 = self.add_folder("ignored2")
        self.add_file("xyz", "foo", path=ignored2)
        folder = self.add_folder("tracked")
        tracked = self.add_file("tracked", "tracked", path=folder)
        self.add_file(".abstraignore", "ignored\nignored2/")
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [tracked, self.abstra_json_path])

    def test_ignore_wildcard(self):
        self.add_file("ignored.ipynb", "foo")
        tracked = self.add_file("tracked", "bar")
        self.add_file(".abstraignore", "*.ipynb")
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [tracked, self.abstra_json_path])

    def test_ignore_abstra_folder(self):
        abstra = self.add_folder(".abstra")
        self.add_file("credentials", "SECRET", path=abstra)
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [self.abstra_json_path])

    def test_ignore_git_folder(self):
        git = self.add_folder(".git")
        self.add_file("somefile", "somecontent", path=git)
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [self.abstra_json_path])

    def test_ignore_abstra_ignore_file(self):
        self.add_file(".abstraignore", "ignored")
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [self.abstra_json_path])

    def test_ignore_git_ignore_file(self):
        self.add_file(".gitignore", "ignored")
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [self.abstra_json_path])



import unittest
from abstra_cli.credentials import get_credentials, delete_credentials, set_credentials
from .fixtures import init_dir, clear_dir


class TestCredentials(unittest.TestCase):
    def setUp(self):
        self.root = init_dir()

    def tearDown(self):
        clear_dir(self.root)

    def test_get_credentials(self):
        self.assertIsNone(get_credentials(self.root))

        set_credentials(self.root, "test")
        self.assertEqual(get_credentials(self.root), "test")

        delete_credentials(self.root)
        self.assertIsNone(get_credentials(self.root))



from ..widget_base import Input
from typing import Optional, Union
from ..response_types import PhoneResponse


class PhoneInput(Input):
    type = "phone-input"

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.label = label
        self.initial_value = kwargs.get("initial_value", "")
        self.required = kwargs.get("required", True)
        self.hint = kwargs.get("hint", None)
        self.placeholder = kwargs.get("placeholder", "")
        self.columns = kwargs.get("columns", 1)
        self.full_width = kwargs.get("full_width", False)
        self.disabled = kwargs.get("disabled", False)

    def json(self, **kwargs):
        return {
            "type": self.type,
            "key": self.key,
            "label": self.label,
            "initialValue": PhoneInput.__revert_value(self.initial_value),
            "placeholder": self.placeholder,
            "required": self.required,
            "hint": self.hint,
            "columns": self.columns,
            "fullWidth": self.full_width,
            "disabled": self.disabled,
        }

    @staticmethod
    def __revert_value(value: Optional[Union[PhoneResponse, str, dict]]):
        if isinstance(value, str):
            return {"countryCode": None, "nationalNumber": value}
        if isinstance(value, dict):
            return {
                "countryCode": value.get("country_code", ""),
                "nationalNumber": value.get("national_number", ""),
            }
        if isinstance(value, PhoneResponse):
            return {
                "raw": value.raw,
                "masked": value.masked,
                "countryCode": value.country_code,
                "nationalNumber": value.national_number,
            }
        return None

    @staticmethod
    def __convert_answer(answer) -> Optional[PhoneResponse]:
        if not answer:
            return None
        return PhoneResponse(
            masked=answer.get("masked"),
            raw=answer.get("raw"),
            country_code=answer.get("countryCode"),
            national_number=answer.get("nationalNumber"),
        )

    def convert_answer(self, answer) -> Optional[PhoneResponse]:
        """
        Returns:
            PhoneResponse: A dict containing the value entered by the user ({"raw": str, "masked": str})
        """
        return self.__convert_answer(answer)



from ..widget_base import Input
from typing import Any


class CustomInput(Input):
    type = "custom-input"

    def __init__(self, key: str, html_body: str, **kwargs):
        super().__init__(key)
        self.html_body = html_body
        self.label = kwargs.get("label", "")
        self.html_head = kwargs.get("html_head", "")
        self.css = kwargs.get("css", "")
        self.js = kwargs.get("js", "")
        self.initial_value = kwargs.get("initial_value", None)
        self.columns = kwargs.get("columns", 1)
        self.full_width = kwargs.get("full_width", False)

    def json(self, **kwargs):
        return {
            "type": self.type,
            "key": self.key,
            "label": self.label,
            "initialValue": self.initial_value,
            "htmlBody": self.html_body,
            "htmlHead": self.html_head,
            "css": self.css,
            "js": self.js,
            "columns": self.columns,
            "fullWidth": self.full_width,
        }

    def convert_answer(self, answer) -> Any:
        return answer



import typing
from dataclasses import dataclass
from .metadata import metadata


@dataclass
class WEvents:
    key: str
    description: str
    payloadSchema: typing.Any


@dataclass
class WDashProperties:
    minWidth: int
    minHeight: int
    initialWidth: int
    initialHeight: int


@dataclass
class WPythonAPIParams:
    argName: str
    description: str
    typeName: str
    isKwarg: bool
    default: typing.Any
    formOnly: bool = False
    dashesInitialValue: typing.Optional[str] = None
    typeDescription: typing.Optional[typing.List[str]] = None


@dataclass
class WPythonAPIReturns:
    typeName: typing.Optional[str]
    typeDescription: str


class WPythonAPI:
    name: str
    params: typing.List[WPythonAPIParams]
    returns: typing.List[WPythonAPIReturns]

    def __init__(self, name, params, returns):
        self.name = name
        self.params = [WPythonAPIParams(**param) for param in params]
        self.returns = [WPythonAPIReturns(**returns) for returns in returns]


@dataclass
class WBrokerAPIParams:
    argName: str
    typeName: typing.Union[str, typing.List[str]]
    description: str
    default: typing.Any = None
    isOptional: bool = False
    formOnly: bool = False
    items: typing.Any = None
    oneOf: typing.Any = None
    properties: typing.Any = None


class WBrokerAPI:
    params: typing.List[WBrokerAPIParams]

    def __init__(self, params):
        self.params = [WBrokerAPIParams(**param) for param in params]


@dataclass
class WExample:
    props: typing.Dict[str, typing.Any]
    name: str
    description: str
    key: str
    code: str


class WMetadata:
    name: str
    description: str
    type: str
    autoHeight: bool
    events: typing.List[WEvents]
    dashProperties: WDashProperties
    pythonAPI: WPythonAPI
    brokerAPI: WBrokerAPI
    examples: typing.List[WExample]
    dashOnly: bool
    formOnly: bool

    def __init__(
        self,
        name,
        description,
        type,
        autoHeight,
        events,
        dashProperties,
        pythonAPI,
        brokerAPI,
        examples,
        dashOnly=False,
        formOnly=False,
    ):
        self.name = name
        self.description = description
        self.type = type
        self.autoHeight = autoHeight
        self.events = [WEvents(**event) for event in events]
        self.dashProperties = WDashProperties(**dashProperties)
        self.pythonAPI = WPythonAPI(**pythonAPI)
        self.brokerAPI = WBrokerAPI(**brokerAPI)
        self.examples = [WExample(**example) for example in examples]
        self.dashOnly = dashOnly
        self.formOnly = formOnly


metadata_objects = {k: WMetadata(**v) for k, v in metadata.items()}



import sys
from ..apis import main as api_main


def get_subdomain():
    query = """
        query Subdomains {
            subdomains {
                name
            }
        }
    """

    subdomains = api_main.hf_hasura_runner(query, {}).get("subdomains", [])
    if not len(subdomains):
        print("Could not find subdomain.")
        sys.exit(1)

    return subdomains[0]["name"]


def update_subdomain(old_name, new_name):
    query = """
        mutation UpdateSubdomain($old_name: String!, $new_name: String!) {
            update_subdomains(where: {name: {_eq: $old_name}}, _set: {name: $new_name}) {
                returning {
                    name
                }
            }
        }
    """
    return (
        api_main.hf_hasura_runner(query, {"old_name": old_name, "new_name": new_name})
        .get("update_subdomains", {})
        .get("returning", [])
    )



import flask, flask_sock, os, sys

from ..api import API
from .utils import send_from_dist
from ..runtimes import run_dash, run_form, run_hook


def get_player_bp(api: API):
    bp = flask.Blueprint("player", __name__)
    sock = flask_sock.Sock(bp)
    if api.root_path not in sys.path:
        sys.path.append(str(api.root_path))

    @bp.route("/_api/<path:id_or_path>", methods=["GET"])
    def get_runner_data(id_or_path):
        runtime = api.get_page_runtime(id_or_path)
        if not runtime:
            print("404", id_or_path)
            flask.abort(404)

        workspace = api.get_workspace()
        res = {}
        res[runtime.runner_type] = {
            **runtime.runner_dto,
            "workspace": workspace.runner_dto,  # TODO: fix this in frontend
        }
        return res

    @bp.route("/_version", methods=["GET"])
    def get_version():
        return os.getenv("ABSTRA_BUILD_ID")

    @sock.route("/_socket")
    def websocket(conn: flask_sock.Server):
        try:
            dash_path = flask.request.args.get("dashPath")
            if dash_path is not None:
                dash = api.get_dash(dash_path)
                if not dash:
                    conn.close(reason=404, message="Not found")
                    return

                code = api.read_text_file(dash.file) if dash.file else ""
                return run_dash(conn, dash, code)

            form_path = flask.request.args.get("formPath")
            if form_path is not None:
                form = api.get_form(form_path)
                if not form:
                    conn.close(reason=404, message="Not found")
                    return

                code = api.read_text_file(form.file) if form.file else ""
                return run_form(conn, form, code)

        finally:
            conn.close(message="Done")

    @bp.route("/_files", methods=["PUT"])
    def _upload_file():
        files = flask.request.files
        if len(files) == 0:
            flask.abort(400)

        paths = []
        for file in files.values():
            paths.append(api.save_file(file))
        return paths

    @bp.route("/_files/<path:path>", methods=["GET"])
    def _get_file(path):
        return flask.send_file(api.get_file(path))

    @bp.route("/_assets/logo", methods=["GET"])
    def _logo():
        logo_path = api.get_workspace().logo_url
        return flask.send_from_directory(directory=api.root_path, path=logo_path)

    @bp.route("/_assets/background", methods=["GET"])
    def _background():
        background_path = api.get_workspace().theme
        return flask.send_from_directory(directory=api.root_path, path=background_path)

    @bp.route("/_hooks/<path:path>", methods=["POST", "GET", "PUT", "DELETE", "PATCH"])
    def hook_runner(path):
        hook = api.get_hook(path)
        if not hook:
            flask.abort(404)

        if not hook.file:
            flask.abort(500)

        code = api.read_text_file(hook.file)  # TODO: handle 404
        body, status, headers = run_hook(flask.request, hook, code)
        return flask.Response(status=status, headers=headers, response=body)

    @bp.route("/<path:filename>", methods=["GET"])
    def spa(filename: str):
        return send_from_dist(filename, "player.html")

    return bp



class Message:
    type: str
    data: dict

    def __init__(self, data: dict):
        self.data = data

    def to_json(self, is_preview: bool):
        return {"type": self.type, **self.data}


class StdioMessage(Message):
    type = "stdio"

    def __init__(self, type: str, log: str):
        super().__init__({"log": log})
        self.type = type


class AuthRequireInfoMessage(Message):
    type = "auth:require-info"

    def __init__(self):
        super().__init__({})


class AuthInvalidJWTMessage(Message):
    type = "auth:invalid-jwt"

    def __init__(self):
        super().__init__({})


class AuthValidJWTMessage(Message):
    type = "auth:valid-jwt"

    def __init__(self):
        super().__init__({})



from .TextareaInput import TextareaInput
from .ListInput import ListInput
from .CustomInput import CustomInput
from .VideoInput import VideoInput
from .TagInput import TagInput
from .PandasOutput import PandasOutput
from .TimeInput import TimeInput
from .RichTextInput import RichTextInput
from .NumberSliderInput import NumberSliderInput
from .CheckboxInput import CheckboxInput
from .CnpjInput import CnpjInput
from .ProgressOutput import ProgressOutput
from .CpfInput import CpfInput
from .TextInput import TextInput
from .PhoneInput import PhoneInput
from .ImageInput import ImageInput
from .ToggleInput import ToggleInput
from .ChecklistInput import ChecklistInput
from .LinkOutput import LinkOutput
from .ImageOutput import ImageOutput
from .HtmlOutput import HtmlOutput
from .EmailInput import EmailInput
from .MultipleChoiceInput import MultipleChoiceInput
from .CodeInput import CodeInput
from .FileOutput import FileOutput
from .DateInput import DateInput
from .MarkdownOutput import MarkdownOutput
from .PasswordInput import PasswordInput
from .PlotlyOutput import PlotlyOutput
from .NumberInput import NumberInput
from .KanbanBoardInput import KanbanBoardInput
from .CardsInput import CardsInput
from .AnswerSheetInput import AnswerSheetInput
from .IframeOutput import IframeOutput
from .NpsInput import NpsInput
from .LatexOutput import LatexOutput
from .CurrencyInput import CurrencyInput
from .FileInput import FileInput
from .TextOutput import TextOutput
from .RatingInput import RatingInput
from .ClickInput import ClickInput
from .DropdownInput import DropdownInput
from .PandasRowSelectionInput import PandasRowSelectionInput



from .vars import Vars
from .jobs import Jobs
from .forms import Forms
from .hooks import Hooks
from .files import Files
from .packages import Packages
from .resources import Resource
from .dashes import Dashes
from .subdomains import Subdomains
from .workspaces import Workspaces



from ..widget_base import Output
from ..file_utils import convert_file
from typing import Union
import io


class ImageOutput(Output):
    type = "image-output"

    def __init__(self, image: Union[str, io.IOBase], **kwargs):
        self.image = image
        self.subtitle = kwargs.get("subtitle", "")
        self.columns = kwargs.get("columns", 1)
        self.full_width = kwargs.get("full_width", False)
        self.label = kwargs.get("label", None)

    def json(self, **kwargs):
        return {
            "type": self.type,
            "imageUrl": convert_file(self.image),
            "subtitle": self.subtitle,
            "columns": self.columns,
            "fullWidth": self.full_width,
            "label": self.label,
        }
